#!/bin/bash
########################################################################
# build_pandora.sh
#
# Build pandora and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_07_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}
thisdir=$(cd "${BASH_SOURCE%/*}" && pwd -P)
trace=maybe_xtrace # Could set to xtrace instead for unconditional.

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# ...

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. ${thisdir}/version_info.sh || \
  ssi_die "unable to load version info from ${thisdir}/version_info"
####################################

##################
# Other useful variables. *C*
pkgtarname=${package}-${PandoraPFA_ver}
pkgtarfile=${pkgtarname}.tar
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e20,e26,c7,c14} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS NO_DEBUG_SOURCE || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit

[ -n "${ROOTSYS}" ] && ssi_info "ROOTSYS: ${ROOTSYS}" || \
  ssi_die "failed to set up ROOT"
####################################

####################################
# Build-only dependencies. *C*
#
${trace} setup_cmake --latest || exit
! (( NO_NINJA )) && ${trace} setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

${trace} mkdir -p "${pkgtarname}" && \
  cd "${pkgtarname}" || ssi_die "unable to source directory ${pkgtarname}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.

# Unpack the top level into the current directory without additional
# subdir.
${trace} prepare_source -e --strip-components,1 \
  "${tardir}/PandoraPFA-${PandoraPFA_ver}.tar" || exit

# Remove download method specifications to use our pre-loaded sources only.
${trace} sed -Ei'' -e '/^[[:space:]]*GIT_(REPOSITORY|TAG)\b/d' CMakeLists.txt || \
  ssi_die "unable to disable build-time source downloads in CMakeLists.txt"

# Unpack remainder of sources, honoring subdir.
for rep in "${repos[@]:1}"; do
  eval rep_ver=\"\${${rep}_ver}\"
  ${trace} prepare_source "${tardir}/${rep}-${rep_ver}.tar" || exit
done
# patch for gcc12
pwd
${trace} patch -b -p0 < ${patchdir}/pandora-v03-16-00.patch

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

declare -a cmake_options cxxflg
(( want_ninja )) && cmake_options+=(-GNinja)

case $cqual in
  e1[79]|e2[06]) cc=gcc; cxx=g++
    cxxflg+=(-std=c++17 -Wno-implicit-fallthrough);;
  e28) cc=gcc; cxx=g++
    cxxflg+=(-std=c++20 -Wno-implicit-fallthrough);;
  c[27]|c14) cc=clang; cxx=clang++
    cxxflg+=(-std=c++17 -D_LIBCPP_ENABLE_CXX17_REMOVED_FEATURES=1);;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac

cmake_options+=(-DCMAKE_BUILD_TYPE=RELEASE
  -DCMAKE_INSTALL_PREFIX="${pkgdir}"
  -DCMAKE_C_COMPILER="${cc}"
  -DCMAKE_CXX_COMPILER="${cxx}"
  -DCMAKE_CXX_FLAGS="${cxxflg[*]}"
  -DPANDORA_MONITORING:BOOL=ON
  -DROOT_DIR="${ROOTSYS}/etc/cmake"
  -DLAR_PANDORA_CONTENT:BOOL=OFF
  -DLC_PANDORA_CONTENT:BOOL=ON
  # As of v3-15-00, documentation generation is broken.
  -DINSTALL_DOC:BOOL=OFF
  -DEXAMPLE_PANDORA_CONTENT:BOOL=OFF
)

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
[[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}

# For ctest:
[[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
  export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*
${trace} cmake "${cmake_options[@]}" "${sourcedir}/${pkgtarname}" || \
    ssi_die "configure failed"
##################

##################
# Build. *C*

# Pandora's build system doesn't pass -v down to the subprojects, so we
# have to take extra measures.
${trace} env ${DEBUG_BUILD:+VERBOSE=1} \
  cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
  ssi_die "build failed"
##################

##################
# Install. *C*
${trace} cmake --install "${blddir}" ${DEBUG_INSTALL:+--verbose} || \
  ssi_die "install failed for ${rep}"

# Manual installation of required CMake file: shouldn't be
# necessary. See also postprocessing step.
install -m 0644 \
  "${sourcedir}/${pkgtarname}/cmakemodules/MacroCheckPackage"*.cmake \
  "${pkgdir}/lib/cmake/" || \
  ssi_die "unable to manually install required CMake files in lib/cmake/"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.

ssi_info "post-processing for relocatability"

# Need to fix a reference to the src directory for CMake source files to
# find their (manually) installed location.
${trace} perlfilt \
  -wpi'' \
  -e 's&\Q'"${pkgtop}/src/${pkgtarname}/cmakemodules/"'\E&'"${pkgdir}/lib/cmake/&" \
  "${pkgdir}"/*.cmake "${pkgdir}"/lib/cmake/*.cmake && \
  ${trace} fix_config_script \
  =PANDORA_FQ_DIR \
  =PANDORA_INC \
  =PANDORA_LIB \
  =ROOTSYS \
  -- \
  "${pkgdir}"/*.cmake "${pkgdir}"/lib/cmake/*.cmake || \
  ssi_die "post-processing failed"
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
